// HBC3.java
// Three phase simulation in java, with GUI-based input using awt.
// Can also use a run-time form developed in Borland JBuilder

package HBCSim;

import java.io.*;
import java.lang.String;
import SimObs.*;
import SimExec.*;

/** Contains the Cs, initialisation and finalisation for a 3 phase simulation
* of the Harassed Booking Clerk problem introduced in Pidd (1998) Computer
* Simulation in MAnagement Science, John Wiley, Chichester.
* Contains the main() function of the java application.
*/
public class HBC3{
  /** The clerks who serve the customers*/
  public static Resource clerks;
  /** The simulation executive to be used to run the simulation*/
  static Executive exec;
  /** An input screen for setting simulation parameters*/
  static HBCForm screen;
  /** A C to simulate serving a personal customer*/
  static BeginServe startServing;
  /** A C to simulate answering the phone*/
  static BeginTalk startTalking;
  /** A static object to represent a single observer*/
  static Observer theObserver;

  /**
  * Constructor that sets up the basic parameters of the simulation.
  * It instantiates an executive, sets up Debug.txt as a trace file, creates
  * 2 clerks as a resource pool, instantiates the first personal customer,
  * first phone call, and the observer.
  * Instantiates the Bs (these are classes via the Activity interface.
  * Reads the parameters from the data input screen, sets up the queues and
  * random number streams, schedules the arrivals of the first personal
  * customer and first phone call and adds the two Cs to the cList.
  */
  public HBC3() {
    exec = new Executive();
    Trace deBug = new Trace("debug.txt");
    clerks = new Resource("clerks", 2);
    PersCust firstCust = new PersCust("PersCust");
    PhoneEnq firstEnq = new PhoneEnq("PhoneEnq");
    theObserver = new Observer("The Observer");
    exec.addNewEntity(firstCust);
    exec.addNewEntity(firstEnq);
    exec.addNewEntity(theObserver);
    startServing = new BeginServe();
    startTalking = new BeginTalk();
    exec.setTrace(true);

    // Get simulation parameters from input screen
    PersCust.setTimes(screen.getPersA(), screen.getPersS());
    PhoneEnq.setTimes(screen.getPhoneA(), screen.getPhoneS());
    exec.setRunDuration(screen.getSimDur());
    Observer.setInterval(screen.getInterval());

    // Establish static variables of customer classes
    PersCust.setUpQs();
    PersCust.seedRND(9876, 5432);
    PhoneEnq.setUpQs();
    PhoneEnq.seedRND(1234, 5678);

    // Schedule first arrivals & observation
    Executive.schedule(firstCust, PersCust.nextArrival, PersCust.getIAT());
    Executive.schedule(firstEnq, PhoneEnq.nextCall, PhoneEnq.getIAT());
    theObserver.show();
    Executive.schedule(theObserver, theObserver.nextObservation, Observer.getInterval());
    theObserver.show();

    // Set up CList
    Executive.addC(startServing);
    Executive.addC(startTalking);

    goForIt();
  }

  /**
  * Runs the simulation
  */
  void goForIt(){
    while (Executive.getClock() <= Executive.getRunDuration()){
      Executive.simulate();
    }
  }

  /**
  * The first C, begin serve, representing the start of the service of a
  * personal customer.
  * Checks whether customers are waiting in the queue and a clerk is free, if so
  * reduces the number of free clerks by 1 and schedules the service to end, using
  * the distribution set up in the PersCust class. Writes to the tracefile.
  */
  class BeginServe implements Activity {
    public void doThisNow(){
      int persServeNum;
      PersCust thisCust;
      while ((PersCust.qIn.size() > 0)&&(clerks.getFreeAmount() > 0)){
        Executive.setCStarted();
        thisCust = (PersCust)PersCust.qIn.takeFromHead();
        Executive.schedule(thisCust, PersCust.serviceOver, PersCust.getServiceTime());
        persServeNum = PersCust.qIn.getNumRemoved();
        clerks.commit(1);
        Trace.traceFile.println("Service " + persServeNum + " starts at " + exec.getClock()
                        + " clerks free: " + clerks.getAmount());
      }
    }
  }

  /**
  * The second C, begin serve, representing the start of the service of a
  * phoner caller.
  * Checks whether calls are waiting in the queue and a clerk is free, if so
  * reduces the number of free clerks by 1 and schedules the conversation to end, using
  * the distribution set up in the PhoneEnq class. Writes to the tracefile.
  */
  class BeginTalk implements Activity {
    public void doThisNow() {
      int phoneConvNum;
      PhoneEnq thisCall;
      while ((PhoneEnq.qIn.size() > 0)&&(clerks.getFreeAmount() > 0)){
        Executive.setCStarted();
        thisCall = (PhoneEnq)PhoneEnq.qIn.takeFromHead();
        Executive.schedule(thisCall, PhoneEnq.callOver, PhoneEnq.getServiceTime());
        phoneConvNum = PhoneEnq.qIn.getNumRemoved();
        clerks.commit(1);
        Trace.traceFile.println("Conversation " + phoneConvNum + " starts at " + exec.getClock()
                        + " clerks free: " + clerks.getAmount());
      }
    }
  }

  /**
  * Could be used to write to the system screen if needed.
  */
  static void DisplayResults(String blurb, int res1, int res2) {
    System.out.println("\n" + blurb + ".. Personal: " + res1 + "   Phone: " + res2);
  }

  /**
  * The main function of the application.
  * Calls up the input data screen.
  */
  public static void main(String args[]) {
    screen = new HBCForm("Harassed booking clerk simulation");
  }
}