//Executive.java
//A Java-based 3 phase simulation system
//M. Pidd

package SimExec;

import SimObs.*;                  // Defines specialist simulation classes

import java.util.*;
import java.lang.*;
import java.io.*;

/**
* Executive class to control a 3 phase simulation in which entity records are
* held in a Vector (details) and from which a temporary Vector (dueList) is
* created at each time beat to contain those entities with events due at the
* current event time. Entities are added to the details Vector before the start
* of an aPhase and unwanted entities are also deleted then.
* Bs are assumed part of application entity classes, represented by integers &
* Cs as belonging to the application alone
* Revised August 2000, consistent syntax
*/
 public class Executive {
  private final static long checkValue = 1073741824;  // A large value
  private static long runDuration = 0;        // Simulation duration
  private static long clock = 0;              // Current clock time
  private static long prevClock = 0;          // Previous clock time
  private static Vector details;              // Entities now in use
  private static Vector dueList;              // Entities due at current B
  private static Vector newEntities;          // List of new entities to be added
  private static Vector toBeKilled;           // Entities to be removed
  private static GEntity currEnt = null;      // Currently active entity
  private static boolean trace;               // To trace or not?
  private static boolean traceEverOn;         // Trace ever been on?
  private static boolean cStarted;            // True if a C started this time
  private static Vector cList;                // Ordered list of active Cs
  private static final long MAXTCELL = 1073741824;   // 2^30

  /**
  * Simple constructor, sets trace and traceEverOn fields to false so that no
  * debug.txt file is created unless it is required.
  */
  public Executive(){
    details = new Vector();
    dueList = new Vector();
    newEntities = new Vector();
    toBeKilled = new Vector();
    cStarted = false;
    cList = new Vector();
    trace = false;
    traceEverOn = false;
  }

  /**
  * Returns with the current value of the simulation time.
  */
  public static long getClock(){
    return clock;
  }

  /**
  * Allows the user to set the planned duration of the simulation.
  */
  public static void setRunDuration(long duration){
        runDuration = duration;
  }

  /**
  * Returns with the intended duration of the simulation
  */
  public static long getRunDuration(){
    return runDuration;
  }

  /**
  * Sets the current value of trace to the indicated boolean state. If this is
  * true, then a debug.txt file will be created. It's probably best only to set
  * this at the start of a simulation run. The variable traceEverOn will be
  * true if the value of trace has ever been true during a simulation run.
  */
  public static void setTrace(boolean thisState) {
    trace = thisState;
    if (trace)
      traceEverOn = true;
  }

  /**
  * Returns with the boolean value of traceEverOn.
  */
  public static boolean getTraceEverOn() {
    return traceEverOn;
  }

  /**
  * Returns with the boolean value of trace.
  */
  public static boolean getTrace() {
    return trace;
  }

  /**
  * Returns with the current entity - that is, the entity whose B is about to be
  * processed or is being processed currently. GEntity is defined as part of the
  * SimObs package.
  */
  public static GEntity getCurrEnt(){
    return currEnt;
  }

  /**
  * Called in user program when an entity is to be destroyed.
  * Places the entity on an temporary Vector, toBeKilled. The entities in
  * toBeKilled are actually killed just before the start of the next aPhase.
  */
  public static void destroyEntity(GEntity thisEnt){
    toBeKilled.addElement(thisEnt);
    Trace.traceFile.println("Added to ToBeKilled");
    for (int loop = 0; loop < toBeKilled.size(); loop++)
      ((GEntity)toBeKilled.lastElement()).show();
  }

  /**
  * Actually destroys the entity.
  * Works down the toBeKilled Vector and removes elements from details Vector.
  * Called just before the aPhase.
  */
  private static void destroy(){
    if (!toBeKilled.isEmpty()){
      for (int loop = 0; loop < toBeKilled.size(); loop++){
        GEntity thisEnt = (GEntity)toBeKilled.elementAt(loop);
        try{
        if (!details.removeElement(thisEnt))
          throw new Error("entity to be killed does not exist");
          details.removeElement(thisEnt);
          Trace.traceFile.print("Trying to remove .. ");
          thisEnt.show();
        }
        catch (Error e){
          System.err.println("ERROR: " + e.getMessage());
          ((GEntity)toBeKilled.elementAt(loop)).show();
          System.exit(99);
        }
      }
      toBeKilled.removeAllElements();
    }
  }

  /**
  * Called in user program when a new entity is to be added.
  * Places the entity on an temporary Vector, NewEntities, from which the new
  * entities are added just before the start of the aPhase.
  */
  public static void addNewEntity(GEntity thisEnt){
    newEntities.addElement(thisEnt);
  }

  /**
  * Actually adds new entities to the details Vector from the NewEntities Vector.
  * Works down the NewEntities Vector and adds elements to the details Vector,
  * called before the aPhase.
  */
  private static void addNew(){
    if (!newEntities.isEmpty()){
      for (int loop = 0; loop < newEntities.size(); loop++)
        details.addElement((GEntity)newEntities.elementAt(loop));
      newEntities.removeAllElements();
    }
  }

  /**
  * Writes current details Vector to traceFile.
  */
  private static void showDetails(){
    Trace.traceFile.println("In showDetails:" + details.size());
    for (int loop = 0; loop < details.size(); loop++)
      ((GEntity)details.elementAt(loop)).show();
  }

  /**
  * Writes current dueList to traceFile
  */
  private static void showDueList(){
    int entNum;
    Trace.traceFile.println("In showDueList:" + dueList.size());
    for (int loop = 0; loop < dueList.size(); loop++){
      entNum = ((Integer)dueList.elementAt(loop)).intValue();
      ((GEntity)details.elementAt(entNum)).show();
    }
  }

  /**
  * Used when an entity needs to be committed to a B at some future time.
  * Checks whether the entity is currently available. If not, this means that it
  * is already committed to a B and so there is a fatal error thrown.
  * Checks whether the interval between the current simulation clock time and
  * the due time of the B is within range, corrects it if not.
  * Calls the commit method of the entity to update its values in the details
  * Vector.
  * Declared as static, permits only a single executive.
  */
  public static void schedule(GEntity thisEnt, Activity nextActivity, long time){
    Trace.traceFile.println("In schedule");
	  try {
      if (!thisEnt.getAvail()){
        throw new Error("Tried to schedule an unavailable entity");
      }
	    Trace.traceFile.println("Commit: " + thisEnt.getName() +
                              ":" + thisEnt.getEntNum());
	    if (time > MAXTCELL - clock)
        time = MAXTCELL;
      else
	      time = clock + time;
      thisEnt.commit(time, nextActivity);
    }
    catch (Error e){
      System.err.println("ERROR: " + e.getMessage());
      thisEnt.show();
      System.exit(99);
    }
  }

  /**
  * Conventional APhase operating on the Details Vector.
  * Throws a fatal error if aList is empty at the start of the aPhase.
  * Destroys any entities in the ToBeKilled Vector
  * then adds any waiting in the NewEntities Vector.
  * Finds those entities with the smallest time cells that are currently
  * unavailable and adds this due now entities to the dueList Vector.
  */
  private static void aPhase(){
    destroy();
    addNew();
    showDetails();
    long minm = checkValue;
    GEntity thisEnt;
    try {
      if (details.isEmpty()){
        throw new Error("details array empty in APhase()");
      }
      for (int entity = 0; entity < details.size(); entity++){
        thisEnt = (GEntity)details.elementAt(entity);
        if(!thisEnt.getAvail()){
          if(thisEnt.getTCell() < minm){
            minm = thisEnt.getTCell();
            dueList.removeAllElements();
            dueList.addElement(thisEnt);
          }
          else if (thisEnt.getTCell() == minm)
            dueList.addElement(thisEnt);
        }
      }
      if ((minm == checkValue)||(minm == 0)){
        throw new Error("Minm value at end of aPhase()");
      }
      clock = minm;
    }
    catch (Error e){
      System.err.println("ERROR: " + e.getMessage());
			System.exit(99);
    }
  }

  /**
  * Conventional B Phase, causing the next Bs of the due now entities to be
  * executed. Does this by working down the dueList and calling each found
  * there to execute its next B. Throws a fatal error if dueList is empty at
  * the start of bPhase.
  */
  private static void bPhase(){
    int numCurrEnts = dueList.size();
    try {
      if (numCurrEnts <= 0){
        throw new Error("dueList empty at start of bPhase");
      }
      for (int loop = 0; loop < numCurrEnts; loop++){
        currEnt = (GEntity)dueList.elementAt(loop);
        currEnt.doNextB();
      }
      dueList.removeAllElements();
    }
    catch (Error e){
      System.err.println("ERROR: " + e.getMessage());
      System.exit(99);
    }
  }

  /**
  * Adds a new C to the end of the CList
  */
  public static void addC(Activity thisC){
    cList.addElement(thisC);
    Trace.traceFile.println(thisC);
  }

  /**
  * Sets CStarted as true to indicate that a C has been executed during the
  * current C phase.
  */
  public static void setCStarted(){
    cStarted = true;
  }

  /**
  * Causes Cs to be attempted until none succeed. Works down cList, firing
  * each C in turn. Any successful C causes cStarted to be set to true, leading
  * to a re-scan.
  */
  public static void cPhase(){
    try {
      if (cList.isEmpty()) {
        throw new Error("CList empty at start of CPhase");
      }
      cStarted = true;
      while (cStarted){
        cStarted = false;
        Trace.traceFile.println("IN C PHASE: " + cList.size());
        for (int loop = 0; loop <= cList.size()-1; loop++){
          ((Activity)cList.elementAt(loop)).doThisNow();
        }
      }
    }
    catch (Error e){
      System.err.println("ERROR: " + e.getMessage());
      System.exit(99);
    }
  }

  /**
  * Executes a complete 3 phase cycle
  */
  public static void simulate() {
    aPhase();
    Trace.traceFile.println("aPhase complete, time = " + clock);
    bPhase();
    cPhase();
  }
}
