// GEntity.java
// M.Pidd

package SimObs;

/**
* General entity class for three phase simulation. Will be used to define the
* entities required in a particular simulation. GEntity needs to be known by
* the Executive class, since it creates several Vectors to contain GEntity
* objects.
*/
public abstract class GEntity {
  /** The name of the entity */
  public String name;
  /** The time cell (time of next state change) of the entity*/
  protected long tCell = 0;
  /** The previous time cell of the entity*/
  protected long prevTime = 0;
  /** If true, the entity is currently not committed to a B*/
  protected boolean avail = true;
  /** Total time for which the entity has been committed thus far*/
  protected long util = 0;
  /** Counter for number of entities created so far*/
  private static int entCounter = 0;
  /** Unique counter ID for this entity*/
  protected int entNum;
  /** The B in which this entity will engage when clock reaches the time cell*/
  protected Activity nextB;

  /**
  * Simple constructor. Gives the entity a name and keeps count of how many
  * entities have been created in total.
  */
  public GEntity(String thisName){
    entCounter++;
    entNum = entCounter;
    name = thisName;
  }

  /**
  * Returns with the name of the entity.
  */
  public String getName() {
    return name;
  }

  /**
  * Returns with the number of the entity.
  */
  public int getEntNum() {
    return entNum;
  }

  /**
  * Returns with the current value of the time cell of the entity.
  */
  public long getTCell(){
    return tCell;
  }

  /**
  * Used to set the value of the time cell of the entity. DANGEROUS!
  */
  private void setTCell(long thisTime) {
    tCell = thisTime;
  }

  /**
  * Returns with a boolean state showing whether the entity is currently
  * committed to a B.
  */
  public boolean getAvail(){
    return avail;
  }

  /**
  * Used to set the value of the avail field of the entity. DANGEROUS!
  */
  private void setAvail(boolean thisState) {
    avail = thisState;
  }

  /**
  * Used to set the next B of the entity. DANGEROUS!
  */
  private void setNextB(Activity nextAct) {
    nextB = nextAct;
  }

  /**
  * Updates the time cell, availability and next B of the entity.
  * Usually called by Executive.schedule after checking that the values passed
  * are sensible. Allows Executive to be recoded if necessary.
  */
  public void commit(long thisTime, Activity nextAct) {
    tCell = thisTime;
    nextB = nextAct;
    util += thisTime;
    avail = false;
  }

  /**
  * Releases an entity by putting avail to false and updating tCell.
  * DANGEROUS unless called from Executive!
  */
  public void release(){
    try {
      if(avail) {
        throw new Error("Tried to release an available entity");
      }
      avail = true;
      prevTime = tCell;
    }
    catch (Error e){
      System.err.println("ERROR: " + e.getMessage());
      show();
      System.exit(99);
    }
  }

  /**
  * Calls the method representing the next B.
  */
  public void doNextB() {
    release();
    nextB.doThisNow();
  }

  /**
  * Returns with the current utilisation of the entity. Utilisation is the total
  * time for which the entity has been committed thus far. This could exceed
  * current simulation clock if the entity is currently scheduled for some B in
  * the distant future.
  */
  public long getUtil(){
    return util;
  }

  /**
  * Writes information about the entity to the tracefile.
  */
  public void show(){
    Trace.traceFile.print(name + ":" + entNum + ", " + nextB + ", @ " + tCell + ", "
                              + avail + " \n");
  }
}
